#include <gint/display.h>
#include <gint/keyboard.h>

#include <stdio.h>
#include <math.h>

#define WIDTH 396
#define HEIGHT 224
#define MAX_REFLECTIONS 1

// needed structs

typedef struct {
    float x, y, z;
} Vector3;

typedef struct {
    Vector3 origin;
    Vector3 direction;
} Ray;

typedef struct {
    Vector3 center;
    float radius;
    Vector3 color;
    float reflection; // Added reflection coefficient
} Sphere;

typedef struct {
    Vector3 position;
    float intensity;
} Light;

typedef struct {
    Vector3 position;
    Vector3 normal;
    Vector3 color;
    float reflection;
} Plane;

// functions

Vector3 subtract(Vector3 a, Vector3 b) {
    Vector3 result;
    result.x = a.x - b.x;
    result.y = a.y - b.y;
    result.z = a.z - b.z;
    return result;
}

float dot(Vector3 a, Vector3 b) {
    return a.x * b.x + a.y * b.y + a.z * b.z;
}


Vector3 scale(Vector3 v, float scalar) {
    Vector3 result;
    result.x = v.x * scalar;
    result.y = v.y * scalar;
    result.z = v.z * scalar;
    return result;
}

Vector3 normalize(Vector3 v) {
    float length = sqrt(v.x * v.x + v.y * v.y + v.z * v.z);
    Vector3 result;
    result.x = v.x / length;
    result.y = v.y / length;
    result.z = v.z / length;
    return result;
}

int intersect(Ray ray, Sphere sphere, float *t) {
    Vector3 oc = subtract(ray.origin, sphere.center);
    float a = dot(ray.direction, ray.direction);
    float b = 2.0 * dot(oc, ray.direction);
    float c = dot(oc, oc) - sphere.radius * sphere.radius;
    float discriminant = b * b - 4 * a * c;
    if (discriminant < 0) {
        return 0;
    } else {
        float root = sqrt(discriminant);
        float t1 = (-b - root) / (2.0 * a);
        float t2 = (-b + root) / (2.0 * a);
        if (t1 > 0.0001) {
            *t = t1;
            return 1;
        } else if (t2 > 0.0001) {
            *t = t2;
            return 1;
        }
    }
    return 0;
}

int intersectPlane(Ray ray, Plane plane, float *t) {
    float denom = dot(plane.normal, ray.direction);
    if (fabs(denom) > 0.0001) {
        Vector3 originToPlane = subtract(plane.position, ray.origin);
        float tValue = dot(originToPlane, plane.normal) / denom;
        if (tValue >= 0) {
            *t = tValue;
            return 1;
        }
    }
    return 0;
}

int intersectScene(Ray ray, Sphere *spheres, int numSpheres, Plane *planes, int numPlanes, float *t, int *sphereIndex, int *planeIndex) {
    int i;
    int hit = 0;
    float closestT = INFINITY;
    for (i = 0; i < numSpheres; i++) {
        float currentT;
        if (intersect(ray, spheres[i], &currentT) && currentT < closestT) {
            closestT = currentT;
            *sphereIndex = i;
            *planeIndex = -1;  // No plane intersection, set to -1
            hit = 1;
        }
    }
    for (i = 0; i < numPlanes; i++) {
        float currentT;
        if (intersectPlane(ray, planes[i], &currentT) && currentT < closestT) {
            closestT = currentT;
            *planeIndex = i;
            *sphereIndex = -1;  // No sphere intersection, set to -1
            hit = 1;
        }
    }
    if (hit) {
        *t = closestT;
    }
    return hit;
}

Vector3 reflect(Vector3 incident, Vector3 normal) {
    return subtract(incident, scale(normal, 2.0 * dot(incident, normal)));
}

Vector3 trace(Ray ray, Sphere *spheres, int numSpheres, Plane *planes, int numPlanes, Light light, int reflections) {
    Vector3 color = {0, 0, 0};
    float t;
    int sphereIndex, planeIndex;
    if (intersectScene(ray, spheres, numSpheres, planes, numPlanes, &t, &sphereIndex, &planeIndex)) {
        if (sphereIndex != -1) {  // Intersection with a sphere
            Vector3 intersectionPoint;
            intersectionPoint.x = ray.origin.x + ray.direction.x * t;
            intersectionPoint.y = ray.origin.y + ray.direction.y * t;
            intersectionPoint.z = ray.origin.z + ray.direction.z * t;
            Vector3 normal = subtract(intersectionPoint, spheres[sphereIndex].center);
            normal = normalize(normal);
            Vector3 lightDirection = subtract(light.position, intersectionPoint);
            lightDirection = normalize(lightDirection);
            float intensity = dot(normal, lightDirection);
            if (intensity > 0) {
                color.x = intensity * spheres[sphereIndex].color.x * light.intensity;
                color.y = intensity * spheres[sphereIndex].color.y * light.intensity;
                color.z = intensity * spheres[sphereIndex].color.z * light.intensity;
            }
            if (reflections < MAX_REFLECTIONS && spheres[sphereIndex].reflection > 0) {
                Vector3 reflectedDir = reflect(ray.direction, normal);
                Ray reflectedRay;
                reflectedRay.origin = intersectionPoint;
                reflectedRay.direction = reflectedDir;
                Vector3 reflectedColor = trace(reflectedRay, spheres, numSpheres, planes, numPlanes, light, reflections + 1);
                color.x += spheres[sphereIndex].reflection * reflectedColor.x;
                color.y += spheres[sphereIndex].reflection * reflectedColor.y;
                color.z += spheres[sphereIndex].reflection * reflectedColor.z;
            }
        } else if (planeIndex != -1) {  // Intersection with a plane
            Vector3 intersectionPoint;
            intersectionPoint.x = ray.origin.x + ray.direction.x * t;
            intersectionPoint.y = ray.origin.y + ray.direction.y * t;
            intersectionPoint.z = ray.origin.z + ray.direction.z * t;
            Vector3 normal = planes[planeIndex].normal;
            Vector3 lightDirection = subtract(light.position, intersectionPoint);
            lightDirection = normalize(lightDirection);
            float intensity = dot(normal, lightDirection);
            if (intensity > 0) {
                color.x = intensity * planes[planeIndex].color.x * light.intensity;
                color.y = intensity * planes[planeIndex].color.y * light.intensity;
                color.z = intensity * planes[planeIndex].color.z * light.intensity;
            }
            if (reflections < MAX_REFLECTIONS && planes[planeIndex].reflection > 0) {
                Vector3 reflectedDir = reflect(ray.direction, normal);
                Ray reflectedRay;
                reflectedRay.origin = intersectionPoint;
                reflectedRay.direction = reflectedDir;
                Vector3 reflectedColor = trace(reflectedRay, spheres, numSpheres, planes, numPlanes, light, reflections + 1);
                color.x += planes[planeIndex].reflection * reflectedColor.x;
                color.y += planes[planeIndex].reflection * reflectedColor.y;
                color.z += planes[planeIndex].reflection * reflectedColor.z;
            }
        }
    }
    return color;
}

int createRGB565Color(int red, int green, int blue) {
    int color = 0;
    color |= (red & 0x1F) << 11;    // 5 bits for red component
    color |= (green & 0x3F) << 5;   // 6 bits for green component
    color |= (blue & 0x1F);         // 5 bits for blue component
    return color;
}

void render(Sphere *spheres, int numSpheres, Plane *planes, int numPlanes, Light light) {
    int x, y, color;
    
    Vector3 cameraPos = {0, 0, -5};
    Vector3 colorvec = {0, 0, 0};
    Vector3 pixelPos;

    Ray ray;

    for (y = 0; y < HEIGHT; y++) {
        for (x = 0; x < WIDTH; x++) {
            pixelPos.x = (2.0 * x - WIDTH) / HEIGHT;
            pixelPos.y = (2.0 * y - HEIGHT) / HEIGHT;
            pixelPos.z = 0;

            ray.origin = cameraPos;
            ray.direction = normalize(subtract(pixelPos, cameraPos));

            colorvec = trace(ray, spheres, numSpheres, planes, numPlanes, light, 0);
            color = createRGB565Color(colorvec.x * 31, colorvec.y * 63, colorvec.z * 31);
            dpixel(x, y, color);
        }
        dupdate();
    }
}

int main() {
    Sphere spheres[] = {
        {{0, 1, 5}, 0.5, {1, 0, 0}, 0.5},     // Red sphere with 50% reflection
        {{0, -1, 5}, 1.2, {0, 0, 1}, 0.5}             // Blue sphere with 50% reflection
    };
    int numSpheres = sizeof(spheres) / sizeof(Sphere);
    Plane planes[] = {
        {{0, 3, 0}, {0, 1, 0}, {0, 1, 0}, 0}  // Green plane with no reflections
    };
    int numPlanes = sizeof(planes) / sizeof(Plane);
    Light light = {{10, 10, -10}, 1};
    dclear(C_WHITE);
    render(spheres, numSpheres, planes, numPlanes, light);
    getkey();
    return 1;
}
